/*
 * OOP class of RGTP connections
 * implements contribution commands
 *   $conn reply itemid grogname text
 *      return is list, one of   OK gsn
 *                               FULL
 *                               CONT itemid lastedit lastreply
 *                               NOITEM reason
 *                               FORMAT TEXT|SUBJECT|GROGNAME reason
 *   $conn continue itemid grogname subject text
 *      return is list, one of   OK itemid gsn
 *                               CONT itemid lastedit lastreply
 *                               NOITEM reason
 *                               FORMAT TEXT|SUBJECT|GROGNAME reason
 *                               TOOLATE
 *   $conn contribute grogname subject text
 *      return is list           OK itemid gsn
 *                               FORMAT TEXT|SUBJECT|GROGNAME reason
 */

#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <limits.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <unistd.h>
#include <ctype.h>

#include "tcl.h"
#include "tclExtend.h"
#include "rgtp.h"

#include "fixup.h"
#include "md5.h"
#include "RG.h"

static const RG_KeyValuePair contriberrors[] = {
  { "423 ", "FORMAT TEXT"     },
  { "424 ", "FORMAT SUBJECT"  },
  { "425 ", "FORMAT GROGNAME" },
  { "410 ", "NOITEM"          },
  {  0                        }
};

int RG_SubmitError(Tcl_Interp *interp, Tcl_DString *dsP,
                   const RG_KeyValuePair *table) {
  const RG_KeyValuePair *search;
  char *p;
  
  p= Tcl_DStringValue(dsP);
  for (search= table; search->key && strncmp(p,search->key,4); search++);
  if (!search->key) return 0;
  strcpy(interp->result,search->value);
  Tcl_AppendElement(interp, p+4);
  Tcl_DStringFree(dsP);
  return 1;
}

int RG_Object_ReplyCmd(Tcl_Interp *interp, RG_Connection *connP,
                        int argc, char **argv) {
  /*
   *   $conn reply itemid grogname text
   *      return is list, one of   OK gsn
   *                               FORMAT TEXT|GROGNAME reason
   *                               NOITEM reason
   *                               FULL
   *                               CONT itemid lastedit lastreply
   */
  int r;
  Tcl_DString ds;
  unsigned long gsn;
  char *p, *ep;
  char buf[20];
  
  if (argc != 4) { interp->result= "wrong # args"; return TCL_ERROR; }
  if (strlen(argv[1]) != ITEMID_LEN) {
    interp->result= "item-id is wrong length"; return TCL_ERROR;
  }
  if (RG_AccessOK(interp,connP,RG_Access_Write) != TCL_OK) return TCL_ERROR;
  r= RG_TransmitData(interp,connP,argv[2],argv[3],contriberrors);
  if (r == TCL_RETURN) return TCL_OK;
  if (r != TCL_OK) return r;
  strcpy(buf,"REPL "); strcat(buf,argv[1]);
  connP->contitem[0]= 0;
  r= RG_CommandResponse(interp,connP,&ds,buf);
  if (r != TCL_OK) return r;
  p= Tcl_DStringValue(&ds);
  if (!strncmp(p,"220 ",4)) {
    gsn= strtoul(p+4,&ep,16);
    if (ep != p+12 || (*ep && !isspace(*ep))) {
      Tcl_AppendResult(interp,
                       "server sent bad 220 reply to REPL:\n", p,
                       (char*)0);
      Tcl_DStringFree(&ds);
      return TCL_ERROR;
    }
    sprintf(interp->result, "OK %08lx", gsn);
    Tcl_DStringFree(&ds);
    return TCL_OK;
  }
  if (RG_SubmitError(interp,&ds,contriberrors)) return TCL_OK;
  if (!strncmp(p,"421 ",4)) {
    strcpy(connP->contitem,argv[1]);
    sprintf(interp->result,"FULL");
    Tcl_DStringFree(&ds);
    return TCL_OK;
  }
  if (!strncmp(p,"122 ",4)) {
    if (strlen(p+4) != 26) {
      Tcl_AppendResult(interp,
                       "server sent bad 122 info re REPL:\n", p,
                       (char*)0);
      Tcl_DStringFree(&ds);
      RG_Violation(connP);
      return TCL_ERROR;
    }
    sprintf(interp->result,"CONT %s",p+4);
    Tcl_DStringFree(&ds);
    r= RG_ReadLineEcho(interp, connP, &ds);
    if (r != TCL_OK) return r;
    if (strncmp(Tcl_DStringValue(&ds),"422 ",4)) {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,
                       "server sent bad response after 122"
                       " info re REPL:\n", p, (char*)0);
      Tcl_DStringFree(&ds);
      RG_Violation(connP);
      return TCL_ERROR;
    }
    Tcl_DStringFree(&ds);
    return TCL_OK;
  }
  Tcl_AppendResult(interp,
                   "server sent bad response to REPL:\n",
                   p, (char*)0);
  Tcl_DStringFree(&ds);
  RG_Violation(connP);
  return TCL_ERROR;
}

int RG_CheckSubject(Tcl_Interp *interp, char *value, char **startp, int *lenp,
                    char *whatcode, char *whatstring, int maxlen) {
  char *p;
  int l;
  
  p= value;
  while (*p && isspace(*p)) p++;
  if (!*p) {
    sprintf(interp->result,"FORMAT %s {%s: empty, not allowed.}",
            whatcode, whatstring);  return TCL_RETURN;
  }
  l= strlen(p);
  while (isspace(p[l-1])) --l;
  if (l > maxlen) {
    sprintf(interp->result,"FORMAT %s {%s: too long - max. %d chars.}",
            whatcode, whatstring, maxlen);  return TCL_RETURN;
  }
  if (strchr(p,'\n')) {
    sprintf(interp->result,"FORMAT %s {%s: contains a line break - not allowed!}",
            whatcode, whatstring);  return TCL_RETURN;
  }
  *startp= p; *lenp= l;
  return TCL_OK;
}
  
int RG_Object_ContributeCmd(Tcl_Interp *interp, RG_Connection *connP,
                            int argc, char **argv) {
  /*
   *   $conn contribute grogname subject text
   *      return is list           OK itemid gsn
   *                               FORMAT TEXT|SUBJECT|GROGNAME reason
   */
  int r;
  Tcl_DString ds;
  unsigned long gsn;
  char *p, *ep;
  int l;
  char buf[TEXTLINE_MAXLEN+10];
  
  if (argc != 4) { interp->result= "wrong # args"; return TCL_ERROR; }
  r= RG_CheckSubject(interp,argv[2],&p,&l, "SUBJECT", "Subject line",TEXTLINE_MAXLEN);
  if (r == TCL_RETURN) return TCL_OK; else if (r != TCL_OK) return r;
  
  if (RG_AccessOK(interp,connP,RG_Access_Write) != TCL_OK) return TCL_ERROR;
  r= RG_TransmitData(interp,connP,argv[1],argv[3],contriberrors);
  if (r == TCL_RETURN) return TCL_OK;
  if (r != TCL_OK) return r;

  strcpy(buf,"NEWI "); memcpy(buf+5,p,l); buf[5+l]= 0;
  connP->contitem[0]= 0;
  r= RG_CommandResponse(interp,connP,&ds,buf);
  if (r != TCL_OK) return r;
  if (RG_SubmitError(interp,&ds,contriberrors)) return TCL_OK;
  p= Tcl_DStringValue(&ds);
  if (strncmp(p,"120 ",4) || strlen(p+4) != 8) {
    Tcl_AppendResult(interp,
                     "server sent bad response to NEWI:\n",
                     p, (char*)0);
    Tcl_DStringFree(&ds);
    RG_Violation(connP);
    return TCL_ERROR;
  }
  strcpy(buf,p+4);
  Tcl_DStringFree(&ds);
  r= RG_ReadLineEcho(interp, connP, &ds);
  if (r != TCL_OK) return r;
  p= Tcl_DStringValue(&ds);
  if (strncmp(p,"220 ",4)) {
    Tcl_AppendResult(interp,
                     "server sent bad response after 120"
                     " info re NEWI:\n", p, (char*)0);
    Tcl_DStringFree(&ds);
    RG_Violation(connP);
    return TCL_ERROR;
  }
  gsn= strtoul(p+4,&ep,16);
  if (ep != p+12 || (*ep && !isspace(*ep))) {
    Tcl_AppendResult(interp,
                     "server sent bad 220 reply to NEWI:\n", p,
                     (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  sprintf(interp->result,"OK %s %08lx",buf,gsn);
  return TCL_OK;
}

int RG_Object_ContinueCmd(Tcl_Interp *interp, RG_Connection *connP,
                            int argc, char **argv) {
  /*
   *   $conn continue itemid grogname subject text
   *      return is list, one of   OK itemid gsn
   *                               CONT itemid lastedit lastreply
   *                               NOITEM reason
   *                               FORMAT TEXT|SUBJECT|REASON reason
   *                               TOOLATE
   */
  int r;
  Tcl_DString ds;
  unsigned long gsn;
  char *p, *ep;
  int l;
  char buf[TEXTLINE_MAXLEN+10];
  
  if (argc != 5) { interp->result= "wrong # args"; return TCL_ERROR; }
  r= RG_CheckSubject(interp,argv[3],&p,&l, "SUBJECT", "Subject line",TEXTLINE_MAXLEN);
  if (r == TCL_RETURN) return TCL_OK; else if (r != TCL_OK) return r;
  if (!connP->contitem[0]) {
    interp->result= "No preceding reply failed with CONT, so continue impossible";
    return TCL_ERROR;
  }
  if (strcmp(connP->contitem,argv[1])) {
    Tcl_AppendResult(interp,
                     "Item to continue (", argv[1],
                     ") does not match item which may be continued (",
                     connP->contitem, ")", (char*)0);
    return TCL_ERROR;
  }
  if (RG_AccessOK(interp,connP,RG_Access_Write) != TCL_OK) return TCL_ERROR;
  r= RG_TransmitData(interp,connP,argv[2],argv[4],contriberrors);
  if (r == TCL_RETURN) return TCL_OK;
  if (r != TCL_OK) return r;

  if (!connP->contitem[0]) { interp->result= "TOOLATE"; return TCL_OK; }
  strcpy(buf,"CONT "); memcpy(buf+5,p,l); buf[5+l]= 0;
  r= RG_CommandResponse(interp,connP,&ds,buf);
  if (r != TCL_OK) return r;
  if (RG_SubmitError(interp,&ds,contriberrors)) return TCL_OK;
  p= Tcl_DStringValue(&ds);
  if (!strncmp(p,"481 ",4)) {
    strcpy(interp->result,"TOOLATE");
    RG_Break(connP);
    connP->status= RG_Status_NeedReconnect;
    return TCL_OK;
  }
  if (!strncmp(p,"122 ",4)) {
    if (strlen(p+4) != 26) {
      Tcl_AppendResult(interp,
                       "server sent bad 122 info re CONT:\n", p,
                       (char*)0);
      Tcl_DStringFree(&ds);
      RG_Violation(connP);
      return TCL_ERROR;
    }
    sprintf(interp->result,"CONT %s",p+4);
    Tcl_DStringFree(&ds);
    r= RG_ReadLineEcho(interp, connP, &ds);
    if (r != TCL_OK) return r;
    if (strncmp(Tcl_DStringValue(&ds),"422 ",4)) {
      Tcl_ResetResult(interp);
      Tcl_AppendResult(interp,
                       "server sent bad response after 122"
                       " info re CONT:\n", p, (char*)0);
      Tcl_DStringFree(&ds);
      RG_Violation(connP);
      return TCL_ERROR;
    }
    Tcl_DStringFree(&ds);
    return TCL_OK;
  }
  if (strncmp(p,"120 ",4) || strlen(p+4) != ITEMID_LEN) {
    Tcl_AppendResult(interp,
                     "server sent bad response to CONT:\n",
                     p, (char*)0);
    Tcl_DStringFree(&ds);
    RG_Violation(connP);
    return TCL_ERROR;
  }
  connP->contitem[0]= 0;
  strcpy(buf,p+4);
  Tcl_DStringFree(&ds);
  r= RG_ReadLineEcho(interp, connP, &ds);
  if (r != TCL_OK) return r;
  p= Tcl_DStringValue(&ds);
  if (strncmp(p,"220 ",4)) {
    Tcl_AppendResult(interp,
                     "server sent bad response after 120"
                     " info re CONT:\n", p, (char*)0);
    Tcl_DStringFree(&ds);
    RG_Violation(connP);
    return TCL_ERROR;
  }
  gsn= strtoul(p+4,&ep,16);
  if (ep != p+12 || (*ep && *ep!=' ')) {
    Tcl_AppendResult(interp,
                     "server sent bad 220 reply to CONT:\n", p,
                     (char*)0);
    Tcl_DStringFree(&ds);
    return TCL_ERROR;
  }
  Tcl_DStringFree(&ds);
  sprintf(interp->result,"OK %s %08lx",buf,gsn);
  return TCL_OK;
}
